@@ -17,9 +17,12 @@ from ray.autoscaler._private.cli_logger import cli_logger, cf
 import ray.autoscaler._private.subprocess_output_util as cmd_output_util
 from ray.autoscaler._private.constants import \
      RESOURCES_ENVIRONMENT_VARIABLE
+from ray.autoscaler._private.event_system import (CreateClusterEvent,
+                                                  global_event_system)
 
 logger = logging.getLogger(__name__)
 
+NUM_SETUP_STEPS = 7
 READY_CHECK_INTERVAL = 5
 
 
@@ -75,7 +78,6 @@ class NodeUpdater:
             process_runner, use_internal_ip, docker_config)
 
         self.daemon = True
-        self.process_runner = process_runner
         self.node_id = node_id
         self.provider = provider
         # Some node providers don't specify empty structures as
@@ -106,9 +108,6 @@ class NodeUpdater:
         self.docker_config = docker_config
 
     def run(self):
-        cli_logger.old_info(logger, "{}Updating to {}", self.log_prefix,
-                            self.runtime_hash)
-
         if cmd_output_util.does_allow_interactive(
         ) and cmd_output_util.is_output_redirected():
             # this is most probably a bug since the user has no control
@@ -124,18 +123,10 @@ class NodeUpdater:
                           "Applied config {}".format(self.runtime_hash)):
                 self.do_update()
         except Exception as e:
-            error_str = str(e)
-            if hasattr(e, "cmd"):
-                error_str = "(Exit Status {}) {}".format(
-                    e.returncode, " ".join(e.cmd))
-
             self.provider.set_node_tags(
                 self.node_id, {TAG_RAY_NODE_STATUS: STATUS_UPDATE_FAILED})
             cli_logger.error("New status: {}", cf.bold(STATUS_UPDATE_FAILED))
 
-            cli_logger.old_error(logger, "{}Error executing: {}\n",
-                                 self.log_prefix, error_str)
-
             cli_logger.error("!!!")
             if hasattr(e, "cmd"):
                 cli_logger.error(
@@ -235,24 +226,18 @@ class NodeUpdater:
 
     def wait_ready(self, deadline):
         with cli_logger.group(
-                "Waiting for SSH to become available", _numbered=("[]", 1, 6)):
+                "Waiting for SSH to become available",
+                _numbered=("[]", 1, NUM_SETUP_STEPS)):
             with LogTimer(self.log_prefix + "Got remote shell"):
-                cli_logger.old_info(logger, "{}Waiting for remote shell...",
-                                    self.log_prefix)
 
                 cli_logger.print("Running `{}` as a test.", cf.bold("uptime"))
                 first_conn_refused_time = None
                 while time.time() < deadline and \
                         not self.provider.is_terminated(self.node_id):
                     try:
-                        cli_logger.old_debug(logger,
-                                             "{}Waiting for remote shell...",
-                                             self.log_prefix)
-
                         # Run outside of the container
                         self.cmd_runner.run(
                             "uptime", timeout=5, run_env="host")
-                        cli_logger.old_debug(logger, "Uptime succeeded.")
                         cli_logger.success("Success.")
                         return True
                     except ProcessRunnerError as e:
@@ -269,18 +254,23 @@ class NodeUpdater:
                         # however threading this configuration state
                         # is a pain and I'm leaving it for later
 
-                        retry_str = str(e)
+                        retry_str = "(" + str(e) + ")"
                         if hasattr(e, "cmd"):
+                            if isinstance(e.cmd, str):
+                                cmd_ = e.cmd
+                            elif isinstance(e.cmd, list):
+                                cmd_ = " ".join(e.cmd)
+                            else:
+                                logger.debug(f"e.cmd type ({type(e.cmd)}) not "
+                                             "list or str.")
+                                cmd_ = str(e.cmd)
                             retry_str = "(Exit Status {}): {}".format(
-                                e.returncode, " ".join(e.cmd))
+                                e.returncode, cmd_)
 
                         cli_logger.print(
                             "SSH still not available {}, "
                             "retrying in {} seconds.", cf.dimmed(retry_str),
                             cf.bold(str(READY_CHECK_INTERVAL)))
-                        cli_logger.old_debug(logger,
-                                             "{}Node not up, retrying: {}",
-                                             self.log_prefix, retry_str)
 
                         time.sleep(READY_CHECK_INTERVAL)
 
@@ -293,6 +283,8 @@ class NodeUpdater:
 
         deadline = time.time() + NODE_START_WAIT_S
         self.wait_ready(deadline)
+        global_event_system.execute_callback(
+            CreateClusterEvent.ssh_control_acquired)
 
         node_tags = self.provider.node_tags(self.node_id)
         logger.debug("Node tags: {}".format(str(node_tags)))
@@ -300,8 +292,12 @@ class NodeUpdater:
         if node_tags.get(TAG_RAY_RUNTIME_CONFIG) == self.runtime_hash:
             # When resuming from a stopped instance the runtime_hash may be the
             # same, but the container will not be started.
-            self.cmd_runner.run_init(
-                as_head=self.is_head_node, file_mounts=self.file_mounts)
+            init_required = self.cmd_runner.run_init(
+                as_head=self.is_head_node,
+                file_mounts=self.file_mounts,
+                sync_run_yet=False)
+            if init_required:
+                node_tags[TAG_RAY_RUNTIME_CONFIG] += "-invalidate"
 
         # runtime_hash will only change whenever the user restarts
         # or updates their cluster with `get_or_create_head_node`
@@ -314,10 +310,7 @@ class NodeUpdater:
             cli_logger.print(
                 "Configuration already up to date, "
                 "skipping file mounts, initalization and setup commands.",
-                _numbered=("[]", "2-5", 6))
-            cli_logger.old_info(logger,
-                                "{}{} already up-to-date, skip to ray start",
-                                self.log_prefix, self.node_id)
+                _numbered=("[]", "2-6", NUM_SETUP_STEPS))
 
         else:
             cli_logger.print(
@@ -327,7 +320,8 @@ class NodeUpdater:
             self.provider.set_node_tags(
                 self.node_id, {TAG_RAY_NODE_STATUS: STATUS_SYNCING_FILES})
             cli_logger.labeled_value("New status", STATUS_SYNCING_FILES)
-            self.sync_file_mounts(self.rsync_up, step_numbers=(2, 6))
+            self.sync_file_mounts(
+                self.rsync_up, step_numbers=(1, NUM_SETUP_STEPS))
 
             # Only run setup commands if runtime_hash has changed because
             # we don't want to run setup_commands every time the head node
@@ -341,11 +335,16 @@ class NodeUpdater:
                 if self.initialization_commands:
                     with cli_logger.group(
                             "Running initialization commands",
-                            _numbered=("[]", 3, 5)):
+                            _numbered=("[]", 4, NUM_SETUP_STEPS)):
+                        global_event_system.execute_callback(
+                            CreateClusterEvent.run_initialization_cmd)
                         with LogTimer(
                                 self.log_prefix + "Initialization commands",
                                 show_status=True):
                             for cmd in self.initialization_commands:
+                                global_event_system.execute_callback(
+                                    CreateClusterEvent.run_initialization_cmd,
+                                    {"command": cmd})
                                 try:
                                     # Overriding the existing SSHOptions class
                                     # with a new SSHOptions class that uses
@@ -369,20 +368,31 @@ class NodeUpdater:
                 else:
                     cli_logger.print(
                         "No initialization commands to run.",
-                        _numbered=("[]", 3, 6))
-                self.cmd_runner.run_init(
-                    as_head=self.is_head_node, file_mounts=self.file_mounts)
+                        _numbered=("[]", 4, NUM_SETUP_STEPS))
+                with cli_logger.group(
+                        "Initalizing command runner",
+                        # todo: fix command numbering
+                        _numbered=("[]", 5, NUM_SETUP_STEPS)):
+                    self.cmd_runner.run_init(
+                        as_head=self.is_head_node,
+                        file_mounts=self.file_mounts,
+                        sync_run_yet=True)
                 if self.setup_commands:
                     with cli_logger.group(
                             "Running setup commands",
                             # todo: fix command numbering
-                            _numbered=("[]", 4, 6)):
+                            _numbered=("[]", 6, NUM_SETUP_STEPS)):
+                        global_event_system.execute_callback(
+                            CreateClusterEvent.run_setup_cmd)
                         with LogTimer(
                                 self.log_prefix + "Setup commands",
                                 show_status=True):
 
                             total = len(self.setup_commands)
                             for i, cmd in enumerate(self.setup_commands):
+                                global_event_system.execute_callback(
+                                    CreateClusterEvent.run_setup_cmd,
+                                    {"command": cmd})
                                 if cli_logger.verbosity == 0 and len(cmd) > 30:
                                     cmd_to_print = cf.bold(cmd[:30]) + "..."
                                 else:
@@ -406,10 +416,14 @@ class NodeUpdater:
                                         "Setup command failed.")
                 else:
                     cli_logger.print(
-                        "No setup commands to run.", _numbered=("[]", 4, 6))
+                        "No setup commands to run.",
+                        _numbered=("[]", 6, NUM_SETUP_STEPS))
 
         with cli_logger.group(
-                "Starting the Ray runtime", _numbered=("[]", 6, 6)):
+                "Starting the Ray runtime", _numbered=("[]", 7,
+                                                       NUM_SETUP_STEPS)):
+            global_event_system.execute_callback(
+                CreateClusterEvent.start_ray_runtime)
             with LogTimer(
                     self.log_prefix + "Ray start commands", show_status=True):
                 for cmd in self.ray_start_commands:
@@ -434,11 +448,10 @@ class NodeUpdater:
                             cli_logger.error("See above for stderr.")
 
                         raise click.ClickException("Start command failed.")
+            global_event_system.execute_callback(
+                CreateClusterEvent.start_ray_runtime_completed)
 
     def rsync_up(self, source, target, docker_mount_if_possible=False):
-        cli_logger.old_info(logger, "{}Syncing {} to {}...", self.log_prefix,
-                            source, target)
-
         options = {}
         options["docker_mount_if_possible"] = docker_mount_if_possible
         options["rsync_exclude"] = self.rsync_options.get("rsync_exclude")
@@ -448,9 +461,6 @@ class NodeUpdater:
                            cf.bold(source), cf.bold(target))
 
     def rsync_down(self, source, target, docker_mount_if_possible=False):
-        cli_logger.old_info(logger, "{}Syncing {} from {}...", self.log_prefix,
-                            source, target)
-
         options = {}
         options["docker_mount_if_possible"] = docker_mount_if_possible
         options["rsync_exclude"] = self.rsync_options.get("rsync_exclude")
