@@ -1,5 +1,6 @@
 from functools import partial
 import gym
+from gym.spaces import Box, Dict, Discrete, MultiDiscrete, Tuple
 import logging
 import numpy as np
 import tree
@@ -11,7 +12,6 @@ from ray.rllib.models.action_dist import ActionDistribution
 from ray.rllib.models.jax.jax_action_dist import JAXCategorical
 from ray.rllib.models.modelv2 import ModelV2
 from ray.rllib.models.preprocessors import get_preprocessor, Preprocessor
-from ray.rllib.models.tf.recurrent_net import LSTMWrapper
 from ray.rllib.models.tf.tf_action_dist import Categorical, \
     Deterministic, DiagGaussian, Dirichlet, \
     MultiActionDistribution, MultiCategorical
@@ -19,14 +19,15 @@ from ray.rllib.models.torch.torch_action_dist import TorchCategorical, \
     TorchDeterministic, TorchDiagGaussian, \
     TorchMultiActionDistribution, TorchMultiCategorical
 from ray.rllib.utils.annotations import DeveloperAPI, PublicAPI
-from ray.rllib.utils.deprecation import DEPRECATED_VALUE
+from ray.rllib.utils.deprecation import DEPRECATED_VALUE, deprecation_warning
 from ray.rllib.utils.error import UnsupportedSpaceException
-from ray.rllib.utils.framework import try_import_tf
+from ray.rllib.utils.framework import try_import_tf, try_import_torch
 from ray.rllib.utils.spaces.simplex import Simplex
 from ray.rllib.utils.spaces.space_utils import flatten_space
 from ray.rllib.utils.typing import ModelConfigDict, TensorType
 
 tf1, tf, tfv = try_import_tf()
+torch, _ = try_import_torch()
 
 logger = logging.getLogger(__name__)
 
@@ -34,14 +35,27 @@ logger = logging.getLogger(__name__)
 # __sphinx_doc_begin__
 MODEL_DEFAULTS: ModelConfigDict = {
     # === Built-in options ===
-    # Number of hidden layers for fully connected net
+    # FullyConnectedNetwork (tf and torch): rllib.models.tf|torch.fcnet.py
+    # These are used if no custom model is specified and the input space is 1D.
+    # Number of hidden layers to be used.
     "fcnet_hiddens": [256, 256],
-    # Nonlinearity for fully connected net (tanh, relu)
+    # Activation function descriptor.
+    # Supported values are: "tanh", "relu", "swish" (or "silu"),
+    # "linear" (or None).
     "fcnet_activation": "tanh",
-    # Filter config. List of [out_channels, kernel, stride] for each filter
+
+    # VisionNetwork (tf and torch): rllib.models.tf|torch.visionnet.py
+    # These are used if no custom model is specified and the input space is 2D.
+    # Filter config: List of [out_channels, kernel, stride] for each filter.
+    # Example:
+    # Use None for making RLlib try to find a default filter setup given the
+    # observation space.
     "conv_filters": None,
-    # Nonlinearity for built-in convnet
+    # Activation function descriptor.
+    # Supported values are: "tanh", "relu", "swish" (or "silu"),
+    # "linear" (or None).
     "conv_activation": "relu",
+
     # For DiagGaussian action distributions, make the second half of the model
     # outputs floating bias variables instead of state-dependent. This only
     # has an effect is using the default fully connected net.
@@ -68,9 +82,44 @@ MODEL_DEFAULTS: ModelConfigDict = {
     # Whether the LSTM is time-major (TxBx..) or batch-major (BxTx..).
     "_time_major": False,
 
+    # == Attention Nets (experimental: torch-version is untested) ==
+    # Whether to use a GTrXL ("Gru transformer XL"; attention net) as the
+    # wrapper Model around the default Model.
+    "use_attention": False,
+    # The number of transformer units within GTrXL.
+    # A transformer unit in GTrXL consists of a) MultiHeadAttention module and
+    # b) a position-wise MLP.
+    "attention_num_transformer_units": 1,
+    # The input and output size of each transformer unit.
+    "attention_dim": 64,
+    # The number of attention heads within the MultiHeadAttention units.
+    "attention_num_heads": 1,
+    # The dim of a single head (within the MultiHeadAttention units).
+    "attention_head_dim": 32,
+    # The memory sizes for inference and training.
+    "attention_memory_inference": 50,
+    "attention_memory_training": 50,
+    # The output dim of the position-wise MLP.
+    "attention_position_wise_mlp_dim": 32,
+    # The initial bias values for the 2 GRU gates within a transformer unit.
+    "attention_init_gru_gate_bias": 2.0,
+    # TODO: Whether to feed a_{t-n:t-1} to GTrXL (one-hot encoded if discrete).
+    # "attention_use_n_prev_actions": 0,
+    # Whether to feed r_{t-n:t-1} to GTrXL.
+    # "attention_use_n_prev_rewards": 0,
+
     # == Atari ==
-    # Whether to enable framestack for Atari envs
-    "framestack": True,
+    # Which framestacking size to use for Atari envs.
+    # "auto": Use a value of 4, but only if the env is an Atari env.
+    # > 1: Use the trajectory view API in the default VisionNets to request the
+    #      last n observations (single, grayscaled 84x84 image frames) as
+    #      inputs. The time axis in the so provided observation tensors
+    #      will come right after the batch axis (channels first format),
+    #      e.g. BxTx84x84, where T=num_framestacks.
+    # 0 or 1: No framestacking used.
+    # Use the deprecated `framestack=True`, to disable the above behavor and to
+    # enable legacy stacking behavior (w/o trajectory view API) instead.
+    "num_framestacks": "auto",
     # Final resized frame dimension
     "dim": 84,
     # (deprecated) Converts ATARI frame to 1 Channel Grayscale image
@@ -95,6 +144,8 @@ MODEL_DEFAULTS: ModelConfigDict = {
     # Deprecated keys:
     # Use `lstm_use_prev_action` or `lstm_use_prev_reward` instead.
     "lstm_use_prev_action_reward": DEPRECATED_VALUE,
+    # Use `num_framestacks` (int) instead.
+    "framestack": True,
 }
 # __sphinx_doc_end__
 # yapf: enable
@@ -153,8 +204,8 @@ class ModelCatalog:
                 "Using custom action distribution {}".format(action_dist_name))
             dist_cls = _global_registry.get(RLLIB_ACTION_DIST,
                                             action_dist_name)
-            dist_cls = ModelCatalog._get_multi_action_distribution(
-                dist_cls, action_space, {}, framework)
+            return ModelCatalog._get_multi_action_distribution(
+                dist_cls, action_space, config, framework)
 
         # Dist_type is given directly as a class.
         elif type(dist_type) is type and \
@@ -163,7 +214,7 @@ class ModelCatalog:
                 MultiActionDistribution, TorchMultiActionDistribution):
             dist_cls = dist_type
         # Box space -> DiagGaussian OR Deterministic.
-        elif isinstance(action_space, gym.spaces.Box):
+        elif isinstance(action_space, Box):
             if len(action_space.shape) > 1:
                 raise UnsupportedSpaceException(
                     "Action space has multiple dimensions "
@@ -179,13 +230,13 @@ class ModelCatalog:
                 dist_cls = TorchDeterministic if framework == "torch" \
                     else Deterministic
         # Discrete Space -> Categorical.
-        elif isinstance(action_space, gym.spaces.Discrete):
+        elif isinstance(action_space, Discrete):
             dist_cls = TorchCategorical if framework == "torch" else \
                 JAXCategorical if framework == "jax" else Categorical
         # Tuple/Dict Spaces -> MultiAction.
         elif dist_type in (MultiActionDistribution,
                            TorchMultiActionDistribution) or \
-                isinstance(action_space, (gym.spaces.Tuple, gym.spaces.Dict)):
+                isinstance(action_space, (Tuple, Dict)):
             return ModelCatalog._get_multi_action_distribution(
                 (MultiActionDistribution
                  if framework == "tf" else TorchMultiActionDistribution),
@@ -198,7 +249,7 @@ class ModelCatalog:
                     "Simplex action spaces not supported for torch.")
             dist_cls = Dirichlet
         # MultiDiscrete -> MultiCategorical.
-        elif isinstance(action_space, gym.spaces.MultiDiscrete):
+        elif isinstance(action_space, MultiDiscrete):
             dist_cls = TorchMultiCategorical if framework == "torch" else \
                 MultiCategorical
             return partial(dist_cls, input_lens=action_space.nvec), \
@@ -213,34 +264,38 @@ class ModelCatalog:
 
     @staticmethod
     @DeveloperAPI
-    def get_action_shape(action_space: gym.Space) -> (np.dtype, List[int]):
+    def get_action_shape(action_space: gym.Space,
+                         framework: str = "tf") -> (np.dtype, List[int]):
         """Returns action tensor dtype and shape for the action space.
 
         Args:
             action_space (Space): Action space of the target gym env.
+            framework (str): The framework identifier. One of "tf" or "torch".
+
         Returns:
             (dtype, shape): Dtype and shape of the actions tensor.
         """
+        dl_lib = torch if framework == "torch" else tf
 
-        if isinstance(action_space, gym.spaces.Discrete):
-            return (action_space.dtype, (None, ))
-        elif isinstance(action_space, (gym.spaces.Box, Simplex)):
-            return (tf.float32, (None, ) + action_space.shape)
-        elif isinstance(action_space, gym.spaces.MultiDiscrete):
-            return (tf.as_dtype(action_space.dtype),
-                    (None, ) + action_space.shape)
-        elif isinstance(action_space, (gym.spaces.Tuple, gym.spaces.Dict)):
+        if isinstance(action_space, Discrete):
+            return action_space.dtype, (None, )
+        elif isinstance(action_space, (Box, Simplex)):
+            return dl_lib.float32, (None, ) + action_space.shape
+        elif isinstance(action_space, MultiDiscrete):
+            return action_space.dtype, (None, ) + action_space.shape
+        elif isinstance(action_space, (Tuple, Dict)):
             flat_action_space = flatten_space(action_space)
             size = 0
             all_discrete = True
             for i in range(len(flat_action_space)):
-                if isinstance(flat_action_space[i], gym.spaces.Discrete):
+                if isinstance(flat_action_space[i], Discrete):
                     size += 1
                 else:
                     all_discrete = False
                     size += np.product(flat_action_space[i].shape)
             size = int(size)
-            return (tf.int64 if all_discrete else tf.float32, (None, size))
+            return dl_lib.int64 if all_discrete else dl_lib.float32, \
+                (None, size)
         else:
             raise NotImplementedError(
                 "Action space {} not supported".format(action_space))
@@ -283,6 +338,8 @@ class ModelCatalog:
                 unflatten the tensor into a ragged tensor.
             action_space (Space): Action space of the target gym env.
             num_outputs (int): The size of the output vector of the model.
+            model_config (ModelConfigDict): The "model" sub-config dict
+                within the Trainer's config dict.
             framework (str): One of "tf2", "tf", "tfe", "torch", or "jax".
             name (str): Name (scope) for the model.
             model_interface (cls): Interface required for the model
@@ -294,6 +351,9 @@ class ModelCatalog:
             model (ModelV2): Model to use for the policy.
         """
 
+        # Validate the given config dict.
+        ModelCatalog._validate_config(config=model_config, framework=framework)
+
         if model_config.get("custom_model"):
             # Allow model kwargs to be overridden / augmented by
             # custom_model_config.
@@ -316,7 +376,23 @@ class ModelCatalog:
                                                      model_interface)
 
             if framework in ["tf2", "tf", "tfe"]:
-                # Track and warn if vars were created but not registered.
+                # Try wrapping custom model with LSTM/attention, if required.
+                if model_config.get("use_lstm") or \
+                        model_config.get("use_attention"):
+                    from ray.rllib.models.tf.attention_net import \
+                        AttentionWrapper
+                    from ray.rllib.models.tf.recurrent_net import LSTMWrapper
+
+                    wrapped_cls = model_cls
+                    forward = wrapped_cls.forward
+                    model_cls = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, LSTMWrapper
+                        if model_config.get("use_lstm") else AttentionWrapper)
+                    model_cls._wrapped_forward = forward
+
+                # Obsolete: Track and warn if vars were created but not
+                # registered. Only still do this, if users do register their
+                # variables. If not (which they shouldn't), don't check here.
                 created = set()
 
                 def track_var_creation(next_creator, **kw):
@@ -345,20 +421,43 @@ class ModelCatalog:
                         # Other error -> re-raise.
                         else:
                             raise e
-                registered = set(instance.variables())
-                not_registered = set()
-                for var in created:
-                    if var not in registered:
-                        not_registered.add(var)
-                if not_registered:
-                    raise ValueError(
-                        "It looks like variables {} were created as part "
-                        "of {} but does not appear in model.variables() "
-                        "({}). Did you forget to call "
-                        "model.register_variables() on the variables in "
-                        "question?".format(not_registered, instance,
-                                           registered))
+
+                # User still registered TFModelV2's variables: Check, whether
+                # ok.
+                registered = set(instance.var_list)
+                if len(registered) > 0:
+                    not_registered = set()
+                    for var in created:
+                        if var not in registered:
+                            not_registered.add(var)
+                    if not_registered:
+                        raise ValueError(
+                            "It looks like you are still using "
+                            "`{}.register_variables()` to register your "
+                            "model's weights. This is no longer required, but "
+                            "if you are still calling this method at least "
+                            "once, you must make sure to register all created "
+                            "variables properly. The missing variables are {},"
+                            " and you only registered {}. "
+                            "Did you forget to call `register_variables()` on "
+                            "some of the variables in question?".format(
+                                instance, not_registered, registered))
             elif framework == "torch":
+                # Try wrapping custom model with LSTM/attention, if required.
+                if model_config.get("use_lstm") or \
+                        model_config.get("use_attention"):
+                    from ray.rllib.models.torch.attention_net import \
+                        AttentionWrapper
+                    from ray.rllib.models.torch.recurrent_net import \
+                        LSTMWrapper
+
+                    wrapped_cls = model_cls
+                    forward = wrapped_cls.forward
+                    model_cls = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, LSTMWrapper
+                        if model_config.get("use_lstm") else AttentionWrapper)
+                    model_cls._wrapped_forward = forward
+
                 # PyTorch automatically tracks nn.Modules inside the parent
                 # nn.Module's constructor.
                 # Try calling with kwargs first (custom ModelV2 should
@@ -399,11 +498,22 @@ class ModelCatalog:
             if not v2_class:
                 raise ValueError("ModelV2 class could not be determined!")
 
-            if model_config.get("use_lstm"):
+            if model_config.get("use_lstm") or \
+                    model_config.get("use_attention"):
+
+                from ray.rllib.models.tf.attention_net import \
+                    AttentionWrapper
+                from ray.rllib.models.tf.recurrent_net import LSTMWrapper
+
                 wrapped_cls = v2_class
                 forward = wrapped_cls.forward
-                v2_class = ModelCatalog._wrap_if_needed(
-                    wrapped_cls, LSTMWrapper)
+                if model_config.get("use_lstm"):
+                    v2_class = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, LSTMWrapper)
+                else:
+                    v2_class = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, AttentionWrapper)
+
                 v2_class._wrapped_forward = forward
 
             # Wrap in the requested interface.
@@ -413,17 +523,32 @@ class ModelCatalog:
 
         # Find a default TorchModelV2 and wrap with model_interface.
         elif framework == "torch":
-            v2_class = \
-                default_model or ModelCatalog._get_v2_model_class(
+            # Try to get a default v2 model.
+            if not model_config.get("custom_model"):
+                v2_class = default_model or ModelCatalog._get_v2_model_class(
                     obs_space, model_config, framework=framework)
-            if model_config.get("use_lstm"):
-                from ray.rllib.models.torch.recurrent_net import LSTMWrapper \
-                    as TorchLSTMWrapper
+
+            if not v2_class:
+                raise ValueError("ModelV2 class could not be determined!")
+
+            if model_config.get("use_lstm") or \
+                    model_config.get("use_attention"):
+
+                from ray.rllib.models.torch.attention_net import \
+                    AttentionWrapper
+                from ray.rllib.models.torch.recurrent_net import LSTMWrapper
+
                 wrapped_cls = v2_class
                 forward = wrapped_cls.forward
-                v2_class = ModelCatalog._wrap_if_needed(
-                    wrapped_cls, TorchLSTMWrapper)
+                if model_config.get("use_lstm"):
+                    v2_class = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, LSTMWrapper)
+                else:
+                    v2_class = ModelCatalog._wrap_if_needed(
+                        wrapped_cls, AttentionWrapper)
+
                 v2_class._wrapped_forward = forward
+
             # Wrap in the requested interface.
             wrapper = ModelCatalog._wrap_if_needed(v2_class, model_interface)
             return wrapper(obs_space, action_space, num_outputs, model_config,
@@ -434,8 +559,6 @@ class ModelCatalog:
             v2_class = \
                 default_model or ModelCatalog._get_v2_model_class(
                     obs_space, model_config, framework=framework)
-            if model_config.get("use_lstm"):
-                raise NotImplementedError("JAXModel's not LSTM-wrappable yet!")
             # Wrap in the requested interface.
             wrapper = ModelCatalog._wrap_if_needed(v2_class, model_interface)
             return wrapper(obs_space, action_space, num_outputs, model_config,
@@ -561,7 +684,7 @@ class ModelCatalog:
     @staticmethod
     def _get_v2_model_class(input_space: gym.Space,
                             model_config: ModelConfigDict,
-                            framework: str = "tf") -> ModelV2:
+                            framework: str = "tf") -> Type[ModelV2]:
 
         VisionNet = None
 
@@ -583,9 +706,13 @@ class ModelCatalog:
                 "framework={} not supported in `ModelCatalog._get_v2_model_"
                 "class`!".format(framework))
 
-        # Discrete/1D obs-spaces.
-        if isinstance(input_space, gym.spaces.Discrete) or \
-                len(input_space.shape) <= 2:
+        # Discrete/1D obs-spaces or 2D obs space but traj. view framestacking
+        # disabled.
+        num_framestacks = model_config.get("num_framestacks", "auto")
+        if isinstance(input_space, (Discrete, MultiDiscrete)) or \
+                len(input_space.shape) == 1 or (
+                len(input_space.shape) == 2 and (
+                num_framestacks == "auto" or num_framestacks <= 1)):
             return FCNet
         # Default Conv2D net.
         else:
@@ -613,4 +740,35 @@ class ModelCatalog:
                 action_space=action_space,
                 child_distributions=child_dists,
                 input_lens=input_lens), int(sum(input_lens))
-        return dist_class
+        return dist_class, dist_class.required_model_output_shape(
+            action_space, config)
+
+    @staticmethod
+    def _validate_config(config: ModelConfigDict, framework: str) -> None:
+        """Validates a given model config dict.
+
+        Args:
+            config (ModelConfigDict): The "model" sub-config dict
+                within the Trainer's config dict.
+            framework (str): One of "jax", "tf2", "tf", "tfe", or "torch".
+
+        Raises:
+            ValueError: If something is wrong with the given config.
+        """
+        if config.get("use_attention") and config.get("use_lstm"):
+            raise ValueError("Only one of `use_lstm` or `use_attention` may "
+                             "be set to True!")
+        if framework == "jax":
+            if config.get("use_attention"):
+                raise ValueError("`use_attention` not available for "
+                                 "framework=jax so far!")
+            elif config.get("use_lstm"):
+                raise ValueError("`use_lstm` not available for "
+                                 "framework=jax so far!")
+
+        if config.get("framestack") != DEPRECATED_VALUE:
+            deprecation_warning(
+                old="framestack", new="num_framestacks (int)", error=False)
+            # If old behavior is desired, disable traj. view-style
+            # framestacking.
+            config["num_framestacks"] = 0
